/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

%{
#include <stdio.h>
#include "epsilon.h"
#include "types.h"
#include "environment.h"
#include "multibuffer.h"
#include "epsilonsyntax.h"

/*IDENTIFIER {LETTER}({LETTER}|{DIGIT}|"_")* */
/*
IDENTIFIER [^\ \n\r()\\\:\.]+
*/
%}

%option yylineno

DIGIT   [0-9]
LOWER   [a-z]
UPPER   [A-Z]
LETTER  {LOWER}|{UPPER}

IDENTIFIER [^0-9\ \r\n()\\\:\.]([^\ \r\n()\\\:\.]*)
SIGN       (\-|\+)
INTEGER    {SIGN}?{DIGIT}+

CESCAPE    \\.|\\{DIGIT}|\\{DIGIT}{DIGIT}|\\{DIGIT}{DIGIT}{DIGIT}
CSTRING    \"([^\n\\"]|{CESCAPE})*\"
STRING     {CSTRING}
POINTED    {DIGIT}+"."{DIGIT}*
FLOAT      {SIGN}?{POINTED}|{POINTED}e{SIGN}{DIGIT}+

WHITESPACE (\ |\t|\n|\r)+
COMMENT    "/*"([^/]|"/"[^*])*"*/"|"//"(.)*\n

%%
{WHITESPACE}|{COMMENT}  {/* Do nothing */}
{INTEGER}     {return INTEGER_CONSTANT;}
{FLOAT}       {return FLOAT_CONSTANT;}
{STRING}      {return STRING_CONSTANT;}

"+"           {return PLUS;}
"-"           {return MINUS;}
"*"           {return TIMES;}
"**"          {return POWER;}
"/"           {return DIVIDED;}
"mod"         {return MODULO;}
"("           {return OPEN_PAR;}
")"           {return CLOSE_PAR;}
"<"           {return LESS;}
">"           {return GREATER;}
"="           {return EQUAL;}
"<="          {return LESS_OR_EQUAL;}
">="          {return GREATER_OR_EQUAL;}
"=/="         {return DIFFERENT;}

";"           {return SEMICOLON;}

true          {return TRUE;}
false         {return FALSE;}
and           {return AND;}
or            {return OR;}
not           {return NOT;}

if            {return IF;}
then          {return THEN;}
else          {return ELSE;}

define        {return DEFINE;}

"@@"          {return CONCATENATE;}

"::"          {return CONS;}
head          {return HEAD;}
tail          {return TAIL;}

\\            {return LAMBDA;}
\.            {return DOT;}

":"           {return COLON;}

let           {return LET;}
letrec        {return LETREC;}
be            {return BE;}
in            {return IN;}

rec           {return REC;}

from          {return FROM;}
to            {return TO;}
"@"           {return AT;}
length        {return LENGTH;}

succ          {return SUCC;}
pred          {return PRED;}

input_integer {return INPUT_INTEGER;}
input_float   {return INPUT_FLOAT;}
input_string  {return INPUT_STRING;}

integer       {return INTEGER;}
float         {return FLOAT;}
string        {return STRING;}
boolean       {return BOOLEAN;}

of            {return OF;}
list          {return LIST;}
array         {return ARRAY;}
promise       {return PROMISE;}

empty         {return EMPTY;}

force         {return FORCE;}
delay         {return DELAY;}

integer_to_float {return INTEGER_TO_FLOAT;}
float_to_integer {return FLOAT_TO_INTEGER;}

"->"          {return ARROW;}

","           {return COMMA;}

"["           {return OPEN_BRACKET;}
"]"           {return CLOSE_BRACKET;}

"|"           {return SELECTOR;}

"postfix"     {return POSTFIX;}

{IDENTIFIER}  {
                type_t type;
                int unused;
                int precedence;
                
                if(lookup_with_precedence(yytext, &unused, &unused, &type, &precedence) == 0){
                  if(is_a_postfix_operator_type(type)){
                    switch(precedence){
                      case 1:
                        return POSTFIX_OPERATOR_1;
                      case 2:
                        return POSTFIX_OPERATOR_2;
                      case 3:
                        return POSTFIX_OPERATOR_3;
                      case 4:
                        return POSTFIX_OPERATOR_4;
                      case 5:
                        return POSTFIX_OPERATOR_5;
                      case 6:
                        return POSTFIX_OPERATOR_6;
                      case 7:
                        return POSTFIX_OPERATOR_7;
                      case 8:
                        return POSTFIX_OPERATOR_8;
                      case 9:
                        return POSTFIX_OPERATOR_9;
                      case 10:
                        return POSTFIX_OPERATOR_10;
                      case 11:
                        return POSTFIX_OPERATOR_11;
                      case 12:
                        return POSTFIX_OPERATOR_12;
                      case 13:
                        return POSTFIX_OPERATOR_13;
                      case 14:
                        return POSTFIX_OPERATOR_14;
                      case 15:
                        return POSTFIX_OPERATOR_15;
                      case 16:
                        return POSTFIX_OPERATOR_16;
                      case 17:
                        return POSTFIX_OPERATOR_17;
                      case 18:
                        return POSTFIX_OPERATOR_18;
                      case 19:
                        return POSTFIX_OPERATOR_19;
                      case 20:
                        return POSTFIX_OPERATOR_20;
		      default:{
                        fprintf(stderr,"(precedence is %i)\n",precedence);
                        yyerror("{IDENTIFIER} in epsilon.l: this cannot happen");
  		      }
		    } /* switch */
                  }
		  else if(is_an_infix_operator_type(type))
                    return INFIX_OPERATOR;
                  else
                    return IDENTIFIER;
		}
                else
                  return IDENTIFIER;
              }

%%

