/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include "term.h"

#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>

#define MAX_ATOM_LENGTH   1000


enum type_of_term_enum{
  constructor_type,
  variable_type
};


struct constructor_struct{
  char name[MAX_ATOM_LENGTH+1];
  int ariety;
} *constructors_table;

#define CONSTRUCTORS_TABLE_INITIAL_SIZE 4
int constructors_table_size;
constructor_t next_constructor;


/* The variables table is only created just before unification
   begins and is destroyed right after its end. */
struct variable_struct{
  char name[MAX_ATOM_LENGTH+1];
} *variables_table; 

#define VARIABLES_TABLE_INITIAL_SIZE 4
int variables_table_size;
variable_t next_variable;


struct term_struct{
  enum type_of_term_enum type_of_term;
  union{
    struct{
      constructor_t constructor;
      term_t *subterms;
    };
    variable_t    variable;
  }; /* union */
  term_t set; /* to implement unification */
} *terms_table;

#define TERMS_TABLE_INITIAL_SIZE 16
int terms_table_size;
term_t next_term;

term_t error; /* a special term */

void init_terms(){
  constructors_table=(struct constructor_struct*)malloc(sizeof(struct constructor_struct)*CONSTRUCTORS_TABLE_INITIAL_SIZE);
  constructors_table_size=CONSTRUCTORS_TABLE_INITIAL_SIZE;
  next_constructor=0;

  variables_table=(struct variable_struct*)malloc(sizeof(struct variable_struct)*VARIABLES_TABLE_INITIAL_SIZE);
  variables_table_size=VARIABLES_TABLE_INITIAL_SIZE;
  next_variable=0;

  terms_table=(struct term_struct*)malloc(sizeof(struct term_struct)*TERMS_TABLE_INITIAL_SIZE);
  terms_table_size=TERMS_TABLE_INITIAL_SIZE;
  next_term=0;

  error=term(constructor("<error>",0));
}

void free_terms(){
  int i;
  
  free(constructors_table);
  free(variables_table);
  for(i=0;i<next_term;i++)
    if(terms_table[i].type_of_term==constructor_type)
      free(terms_table[i].subterms);
  free(terms_table);
}

constructor_t constructor(char* name, int ariety){
  constructor_t r;

  if(next_constructor == constructors_table_size){
    constructors_table_size *= 2;
    constructors_table=
      (struct constructor_struct*)realloc(constructors_table,constructors_table_size*sizeof(struct constructor_struct));
  }
  r=next_constructor++;
  strncpy(constructors_table[r].name,name,MAX_ATOM_LENGTH);
  constructors_table[r].name[MAX_ATOM_LENGTH]='\0';
  constructors_table[r].ariety=ariety;

  return r;
}

/* This is only a test. To do: use an hash table. */
variable_t lookup_variable(char* name){
  variable_t i;  

  for(i=0;i<next_variable;i++)
    if(!strcmp(variables_table[i].name,name))
      return i;

  return -1;
}

/* This is only a test. To do: use an hash table. */
term_t lookup_variable_term(variable_t v){
  term_t i;

  for(i=0;i<next_term;i++)
    if(terms_table[i].type_of_term==variable_type)
      if(terms_table[i].variable==v)
        return i;

  return -1;
}

variable_t variable(char* name){
  variable_t r;

  /* If we already have this variable then return the one we stored. */
  if((r=lookup_variable(name)) != -1)
    return r;

  if(next_variable == variables_table_size){
    variables_table_size *= 2;
    variables_table=(struct variable_struct*)realloc(variables_table,variables_table_size*sizeof(struct variable_struct));
  }
  r=next_variable++;
  strncpy(variables_table[r].name,name,MAX_ATOM_LENGTH);
  variables_table[r].name[MAX_ATOM_LENGTH]='\0';

  return r;
}

variable_t new_variable(char* base){
  static int n=1;
  char name[10];

  sprintf(name,"%s%i",base,n++);

  return variable(name);
}

term_t vterm(variable_t v){
  variable_t r;

  /* If a term made up by this variable is already present then use the one we created before: */
  if((r=lookup_variable_term(v)) != -1)
    return r;

  if(next_term == terms_table_size){
    terms_table_size *= 2;
    terms_table=(struct term_struct*)realloc(terms_table,terms_table_size*sizeof(struct term_struct));
  }
  r=next_term++;
  terms_table[r].type_of_term=variable_type;
  terms_table[r].variable=v;
  terms_table[r].set=-1;

  return r;
}

term_t term(constructor_t c, ...){
  va_list variadics;
  term_t r;
  int i;

  va_start(variadics,c);

  if(next_term == terms_table_size){
    terms_table_size *= 2;
    terms_table=(struct term_struct*)realloc(terms_table,terms_table_size*sizeof(struct term_struct));
  }
  r=next_term++;
  terms_table[r].type_of_term=constructor_type;
  terms_table[r].constructor=c;
  terms_table[r].subterms=(term_t*)malloc(sizeof(term_t)*constructors_table[c].ariety);
  terms_table[r].set=-1;

  for(i=0;i<constructors_table[c].ariety;i++)
    terms_table[r].subterms[i]=va_arg(variadics,term_t);

  va_end(variadics);
  return r;
}

/* Only used in the unification algorithm: */
term_t skeletal_term(constructor_t c){
  term_t r;

  if(next_term == terms_table_size){
    terms_table_size *= 2;
    terms_table=(struct term_struct*)realloc(terms_table,terms_table_size*sizeof(struct term_struct));
  }
  r=next_term++;
  terms_table[r].type_of_term=constructor_type;
  terms_table[r].constructor=c;
  terms_table[r].subterms=(term_t*)malloc(sizeof(term_t)*constructors_table[c].ariety);
  terms_table[r].set=-1;
  /* subterms are not initialized */

  return r;
}

void output_term(FILE* f,term_t t){
  switch(terms_table[t].type_of_term){
    case variable_type:{
      fprintf(f,"%s",variables_table[terms_table[t].variable].name);

      //fprintf(f,"<%i>",terms_table[t].variable);
      break;
    }
    
    case constructor_type:{
      int i;

      fprintf(f,"%s",constructors_table[terms_table[t].constructor].name);
      if(constructors_table[terms_table[t].constructor].ariety!=0){
        fprintf(f,"("); 
        for(i=0;i<constructors_table[terms_table[t].constructor].ariety;i++){
          output_term(f,terms_table[t].subterms[i]);
          if(i!=constructors_table[terms_table[t].constructor].ariety-1)
            fprintf(f,", ");
        } /* for */
        fprintf(f,")");
      } /* if */
      //fprintf(f,"<%i>",t);
      
      break;  
    }

    default:{
      fprintf(stderr,"output_term(): This cannot happen\n");
      exit(EXIT_FAILURE);
    }
  } /* switch */

  /*
  if(terms_table[t].set!=-1){
    fprintf(f,"<");output_term(f,terms_table[t].set);fprintf(f,">");
  }
  */
}

term_t subterm(term_t t,int n){
  return terms_table[t].subterms[n];
}

term_t find(term_t x){
  if(terms_table[x].set == -1)
    return x;
  else
    return find(terms_table[x].set);
}

int is_nullary_constructor(term_t x){
  if(terms_table[x].type_of_term != constructor_type)
    return 0;
  else
    return constructors_table[terms_table[x].constructor].ariety == 0;
}

int are_same_nullary_constructor(term_t x, term_t y){
  if((!is_nullary_constructor(x))||(!is_nullary_constructor(y)))
    return 0;
  else 
    return terms_table[x].constructor == terms_table[y].constructor;
}

int is_nonnullary_constructor(term_t x){
  if(terms_table[x].type_of_term != constructor_type)
    return 0;
  else
    return constructors_table[terms_table[x].constructor].ariety > 0;
}

int are_same_nonnullary_constructor(term_t x, term_t y){
  if((!is_nonnullary_constructor(x))||(!is_nonnullary_constructor(y)))
    return 0;
  else
    return terms_table[x].constructor == terms_table[y].constructor;
}

int is_variable(term_t x){
  return terms_table[x].type_of_term == variable_type;
}

/* Adapted from [Aho, Sethi, Ullman]: p.378 and succ. */

/* Merge the equivalence classes containing x and y: */ 
term_t union_function(term_t m, term_t n){
  term_t rep_m=find(m),
         rep_n=find(n);

  if(!is_variable(rep_m))
    return terms_table[rep_n].set=rep_m;
  else
    // If n is a non-variable I must choose n, else both n and m can do; so I always use n.
    return terms_table[rep_m].set=rep_n;
}

term_t recursive_representative(term_t t_){
  term_t t=find(t_);

  if(is_variable(t))
    return t;
  else if(is_nullary_constructor(t))
    return t;
  else{ /* t is a non-nullary constructor */
    term_t r=skeletal_term(terms_table[t].constructor);
    int i;

    for(i=0;i<constructors_table[terms_table[t].constructor].ariety;i++)
      terms_table[r].subterms[i]=recursive_representative(terms_table[t].subterms[i]);
    
    return r;
  }
}

term_t unify_after_initialization(term_t m, term_t n){
  /* Find representative nodes for m and n: */
  term_t s=find(m),
         t=find(n);

  if((s == t)||are_same_nullary_constructor(s,t))
    return s;
  else if(are_same_nonnullary_constructor(s,t)){
    int i;
    term_t r;
  
    /* Apply 'union': */
    union_function(s,t);    

    /* Try unifying the pairs of subterms: return non-error iff they all match. */
    r=skeletal_term(terms_table[s].constructor);
    for(i=0;i<constructors_table[terms_table[s].constructor].ariety;i++)
      if((terms_table[r].subterms[i] = 
	  /*recursive_representative*/(unify_after_initialization(terms_table[s].subterms[i],
                                                                  terms_table[t].subterms[i]))) == error)
        return error;

    return r; /* Ok, subterms unifications succeeded. */
  }
  else if(is_variable(s)||is_variable(t)){
    /* Ok, match a variable with something: */
    return union_function(s,t);
  }
  else
    /* interior nodes with different constructors */
    return error;    
}

void initialize(term_t x){
  int i;

  terms_table[x].set = -1;
  if(terms_table[x].type_of_term == constructor_type)
    for(i=0;i<constructors_table[terms_table[x].constructor].ariety;i++)
      initialize(terms_table[x].subterms[i]);
}

term_t unify(term_t m, term_t n){
  initialize(m);
  initialize(n);
  return recursive_representative(unify_after_initialization(m,n));
}

int are_terms_equal(term_t x,term_t y){
  if(x==y)
    return 1;
  if(terms_table[x].type_of_term != terms_table[x].type_of_term)
    return 0;

  if(terms_table[x].type_of_term == variable_type)
    // x and y are both variables
    return terms_table[x].variable==terms_table[y].variable; // variables only appear once in the table
  else{
    // x and y are both contructors with parameters
    int i;
    for(i=0;i<constructors_table[terms_table[x].constructor].ariety;i++)
      if(!are_terms_equal(terms_table[x].subterms[i],terms_table[y].subterms[i]))
        return 0;
    return 1;
  }
}

int is_term_more_general_than(term_t m,term_t n){
  return 0; // To do
}

int are_terms_equal_excluding_variables(term_t m, term_t n){
  return 0; // To do
}

/*/
int main(){
  init_terms();

  constructor_t function_c=constructor("function",2);
  constructor_t pair_c=constructor("pair",2);
  constructor_t list_c=constructor("list",1);
  constructor_t int_c=constructor("int",0);
  constructor_t bool_c=constructor("bool",0);

  term_t x_t=vterm(new_variable("type"));
  term_t y_t=vterm(new_variable("type"));
  term_t z_t=vterm(new_variable("type"));
  
  term_t t1=term(function_c,
                  term(list_c,
                       term(bool_c)),
                  term(pair_c,
                       term(bool_c),
                       y_t));
  term_t t2=term(function_c,
                  term(list_c,
                       x_t),
                  z_t);

  term_t t3=term(function_c,x_t,y_t);
  term_t t4=term(function_c,y_t,term(int_c));
 
  output_term(stdout,t3);
  printf("\n<=\n");
  output_term(stdout,t4);
  printf("\n--------------------------------------------------------------\n");
  output_term(stdout,unify(t3,t4));
  //output_term(stdout,unify(t3,t4));
  printf("\n");
 
  free_terms();
}
/**/


