/* This file is part of epsilon, a functional language implementation

epsilon v0.01
Copyright (C) Luca Saiu, 2002

This is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public
License along with this file; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <assert.h>
#include <math.h>

#include "lvm.h"
#include "../as/LVM.h"
#include "../as/LVMsyntax.h"
#include "input.h"
#include "map.h"
#include "../version.h"

void* safe_malloc(size_t size){
  void* r=malloc(size);

#ifdef CHECK_MALLOC
  if(r!=NULL)
    return r;
  else{
    fprintf(stderr,"malloc() failed allocating memory.\n");
    exit(EXIT_FAILURE);
  }  
#else
  return r;
#endif
}

void* safe_realloc(void* current,size_t size){
  void* r=realloc(current,size);

#ifdef CHECK_MALLOC
  if(size==0) /* realloc(X,0) is equivalent to free(X), but what should it return? The
                 documentation says nothing about this question. I choose NULL. */
    return NULL;
  if(r!=NULL)
    return r;
  else{
    fprintf(stderr,"realloc() failed allocating memory.\n");
    exit(EXIT_FAILURE);
  }
#else
  return r;
#endif
}

void fatal(const char *message){
  fprintf(stderr,"\aLVM fatal error: %s.\n",message);
  exit(EXIT_FAILURE);
}

struct LVM_instruction{
  int opcode;
  int argument;
};

struct array_object;

struct LVM{
  void** stack;
  int stack_pointer;
  int stack_size;

  void** constants_stack;
  int constants_stack_pointer;
  int constants_stack_size;

  int frame_pointer;
  int instruction_pointer;

  struct LVM_instruction* instructions;
  int* labels2instruction_number; /* Only used at load time */
  long instructions_number;
};

struct base_object{
  int pins_number;
  enum type_ID type;
};
struct integer_object{
  int pins_number;
  enum type_ID type;
  int value;
};
struct float_object{
  int pins_number;
  enum type_ID type;
  double value;
};
struct string_object{
  int pins_number;
  enum type_ID type;
  char* value;
  size_t size; /* this does not include the trailing '\0' */
};
struct array_object{
  int pins_number;
  enum type_ID type;
  void** value;
  size_t size;
};

/* Allocate storage on the heap and set the pins number to 1;
   this DOES NOT initialize to any meaningful value: */
void* allocate_storage(size_t s){
  void* r=(void*)safe_malloc(s);
  ((struct base_object*)r)->pins_number=0; /* No pins by now */

//  fprintf(stderr,"allocate_storage()\n");
  return r;
}

/* Assign a given value to an existing object: */
void assign_integer(void* object, const int x){
  ((struct integer_object*)object)->value=x;
}

/* Assign a given value to an existing object: */
void assign_float(void* object, const double x){
  ((struct float_object*)object)->value=x;
}

/* Assign a given value to an existing object: */
void assign_string(void* object, const char* x){
  size_t size=strlen(x);

  ((struct string_object*)object)->value=(char*)
               safe_realloc(((struct string_object*)object)->value,
                       sizeof(char)*(size+1));
  ((struct string_object*)object)->size=size;
  strcpy(((struct string_object*)object)->value,x);
}

/* Declarations: */
void pin_object(void*);
void unpin_object(void*);
void* clone_object(void*);

/* Assign a given value to an existing object: */
void assign_array(void* object1, void *object2){
  /* Unpin every non-NULL element of object1: */
  int i;
  for(i=0;i<((struct array_object*)object1)->size;i++)
    unpin_object(((struct array_object*)object1)->value[i]);

  /* Resize object1: */
  ((struct array_object*)object1)->size=((struct array_object*)object2)->size;
  ((struct array_object*)object1)->value=
                      (void**)safe_realloc(((struct array_object*)object1)->value,
                                      ((struct array_object*)object2)->size);
  
  /* Clone every element of object2 into object1 and pin every clone: */
  for(i=0;i<((struct array_object*)object1)->size;i++)
    if(((struct array_object*)object2)->value[i]==NULL)
      ((struct array_object*)object1)->value[i]=NULL;
    else{
      ((struct array_object*)object1)->value[i]=
                  clone_object(((struct array_object*)object2)->value[i]);
      pin_object(((struct array_object*)object1)->value[i]);
    }
}

int* integer2value(void*);
double* float2value(void*);
char* string2value(void*);

void assign_object(void* object1, void* object2){
  /* Verify the correctness of the type: */
  if(((struct base_object*)object1)->type!=
     ((struct base_object*)object2)->type){
    fprintf(stderr,"assign_object(): type mismatch\n");
    exit(EXIT_FAILURE);
  }

  switch(((struct base_object*)object1)->type){
    case integer_type:{
      assign_integer(object1,*integer2value(object2));
      break;
    } /* case integer_type */
    case float_type:{
      assign_float(object1,*float2value(object2));
      break;
    } /* case float_type */
    case string_type:{
      assign_string(object1,string2value(object2));
      break;
    } /* case string_type */
    case array_type:{
      assign_array(object1,object2);
      break;
    } /* case array_type */
    default:{
      fprintf(stderr,"assign_object(): unknown type\n");
      exit(EXIT_FAILURE);
    } /* default */
  } /* switch */
}

/* Allocate and initialize to 0 an integer object: */
void* allocate_integer(){
  void *r=allocate_storage(sizeof(struct integer_object));
  ((struct base_object*)r)->type=integer_type;
  /*assign_integer(r,0); //Don't initialize */

  return r;
}

/* Allocate and initialize to 0.0 a float object: */
void* allocate_float(){
  void *r=allocate_storage(sizeof(struct float_object));
  ((struct base_object*)r)->type=float_type;
  /*assign_float(r,0.0); //Don't initialize */

  return r;
}

/* Allocate and initialize to "" a string object: */
void* allocate_string(){
  void *r=allocate_storage(sizeof(struct string_object));
  ((struct base_object*)r)->type=string_type;
  ((struct string_object*)r)->size=0;
  ((struct string_object*)r)->value=(char*)safe_malloc(sizeof(char));
  strcpy(((struct string_object*)r)->value,"");

  return r;
}

/* Allocate and initialize to an array of NULLs an array object: */
void* allocate_array(size_t s){
  void *r=allocate_storage(sizeof(struct array_object));
  int i;
  
  ((struct base_object*)r)->type=array_type;
  ((struct array_object*)r)->size=s;
  ((struct array_object*)r)->value=(void**)safe_malloc(sizeof(void*)*s);

  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //for(i=0;i<s;i++)
  //  (((struct array_object*)r)->value)[i]=NULL;
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  return r;
}

/* Extract the value from an object: */
int* integer2value(void* object){
  return &(((struct integer_object*)object)->value);
}

/* Extract the value from an object: */
double* float2value(void* object){
  return &(((struct float_object*)object)->value);
}

/* Extract the value from an object: */
char* string2value(void* object){
  return (((struct string_object*)object)->value);
}

/* Extract an element from an array object: */
void* array2value(void* object,int element){
  return ((struct array_object*)object)->value[element];
}

/* Destroy an allocated object, without counting the pins: */
void destroy_integer(struct integer_object* object){
  free(object);
}

/* Destroy an allocated object, without counting the pins: */
void destroy_float(struct float_object* object){
  free(object);
}

/* Destroy an allocated object, without counting the pins: */
void destroy_string(struct string_object* object){
  free(object->value); /* free the buffer of characters */
  free(object);
}

void pin_object(void* object);
void unpin_object(void* object);

/* Destroy an allocated object, without counting the pins.  
   This also works with circular objects: unpinning again an object which
   we are destroying sets the pins number to -1, so destroy_object() is not
   called again and no object is destroyed twice. See also unpin_object() and 
   run_LVM, in case CALL_STACK:.
   Remember NOT to count circular references with pins not to create garbage. */
void destroy_array(struct array_object* object){
  int i;
  
  /* Unpin every object referenced by the array; this will invoke
     destroy_object() for every (even indirectly) linked object where
     the pins' number falls to 0: */
  for(i=0;i<object->size;i++)
    unpin_object(object->value[i]); /* this also works for NULL elements */

  /* Free storage for the array: */
  free(object->value);

  /* Free storage for the object: */
  free(object);
}

/* Destroy an allocated object, without counting the pins: */
void destroy_object(void *object){
//  fprintf(stderr,"destroy_object(%p): ",object);
  switch(((struct base_object*)object)->type){
    case integer_type:{
//      fprintf(stderr,"destroy_object(%i)\n",((struct integer_object*)object)->value);
      destroy_integer((struct integer_object*)object);
      break;
    }
    case float_type:{
//      fprintf(stderr,"destroy_object(%f)\n",((struct float_object*)object)->value);
      destroy_float((struct float_object*)object);
      break;
    }
    case string_type:{
//      fprintf(stderr,"destroy_object(%s)\n",((struct string_object*)object)->value);
      destroy_string((struct string_object*)object);
      break;
    }
    case array_type:{
//      fprintf(stderr,"destroy_object(array[%u])\n",((struct array_object*)object)->size);
      destroy_array((struct array_object*)object);
      break;
    }
    default:{
      fprintf(stderr,"destroy_object(): this cannot happen!\n");
      fprintf(stderr,"                  type==%i\n",
              ((struct base_object*)object)->type);
      exit(EXIT_FAILURE);
    }
  } /* switch */
}

/* Increment the pins' number for an object: */
void pin_object(void* object){
  /*
  if(object != NULL)
    ((struct base_object*)object)->pins_number++;
  return;
  */
  /**/
  if(object != NULL){
    if(is_part_of_a_circular_closure(object))
      if(is_the_environment_part_of_a_circular_closure((struct base_object*)object)){
        ((struct array_object*)(((struct array_object*)object)->value[1]))->pins_number++;
        return;
      }
    ((struct base_object*)object)->pins_number++;
  }
  /**/
}

/* Decrement the pins' number for an object, and delete the object if it
   has no pins left. Do nothing if the object is null. */
void unpin_object(void* object){
//  fprintf(stderr,"unpin_object(%p)\n",object);
  if(object==NULL)
    return;

  /* object is not NULL. */

  /* Circular closures are particular and must be dealt with in a special way: */
  if(is_part_of_a_circular_closure(object)){
    /* This branch should occour more frequently than the other one. */
    //printf("\t\t\t<<<<< UNPINNING PART OF A CLOSURE: ");
    //printf("%i -> ",((struct base_object*)object)->pins_number);

    if(is_the_environment_part_of_a_circular_closure(object)){
      // printf("%i ...",((struct base_object*)(((struct array_object*)object)->value[1]))->pins_number - 1);
      
      if(--(((struct base_object*)(((struct array_object*)object)->value[1]))->pins_number) == 0){
        unpin_object(((struct array_object*)object)->value[0]);
        free((((struct array_object*)(((struct array_object*)object)->value[1])))->value[0]); /* integer */

        free(((struct array_object*)(((struct array_object*)object)->value[1]))->value);
        free(((struct array_object*)object)->value[1]);

        free(((struct array_object*)object)->value);
        free(object);

        //printf(" DESTROYED (ENV) >>>>>\n");
      }
      else
        ;//printf(" OK (ENV) >>>>>\n");
    } /* if branch */
    else{
      /* object is the "closure" part of a closure: */
      //printf("%i ...",((struct base_object*)object)->pins_number - 1);

      if(--(((struct base_object*)object)->pins_number) == 0){
        unpin_object((((struct array_object*)(((struct array_object*)object)->value[1])))->value[0]);
        free(((struct array_object*)object)->value[0]); /* integer */

        free(((struct array_object*)(((struct array_object*)object)->value[1]))->value);
        free(((struct array_object*)object)->value[1]);

        free(((struct array_object*)object)->value);
        free(object);
	//printf(" DESTROYED (CLOS) >>>>>\n");
      }
      else
        ;//printf(" OK (CLOS) >>>>>\n");
    } /* else branch */
    
    return; /* Ok, we have finished with this circular closure. */
  } /* outer if */
;

  /* object is not a circular closure. */

  ((struct base_object*)object)->pins_number--;
  /* Destroy only if the pins are 0; if they are less then 0 then don't destroy; this
     trick is needed for circular objects: it's very important that circular references
     are NOT counted with pins, not to create garbage. */
  if(((struct base_object*)object)->pins_number==0)
    destroy_object(object);
}

/* Push an existing object on the top of the stack of an LVM: */
void push_object(void* object, struct LVM* lvm);

/* Push an existing object on the top of the constants stack of an LVM: */
void push_constant_object(void* object, struct LVM* lvm);

/* Pop the stack of an LVM, unpinning the former top: */
void pop(struct LVM* lvm){
#ifdef DYNAMIC_STACK
  /* Resize the stack if needed: */
  if(lvm->stack_pointer % 100 == 0) /* not to test the complex condition
                                       every time */
  if(lvm->stack_pointer < lvm->stack_size - BLOCK_SIZE*2){
#ifdef STACK_RESIZE_LOG      
  fprintf(stderr,"[POSSO RIDURRE LO STACK: %i -> %i]\n",
          lvm->stack_size,
          lvm->stack_size-BLOCK_SIZE);
#endif
    lvm->stack_size-=BLOCK_SIZE;
    lvm->stack=(void**)safe_realloc((void*)(lvm->stack),
                               sizeof(void*)*(size_t)(lvm->stack_size));
  }
#endif

  /* Unpin the object on the top and decrement the stack pointer: */
  unpin_object(lvm->stack[--(lvm->stack_pointer)]);
}

/* Create an object, initialize it and push on the stack of an LVM: */
void push_new_integer(const int x, struct LVM* lvm){
  void* object=allocate_integer();
  
  assign_integer(object,x);
  push_object(object,lvm);
}

/* Return a newly-created integer object with given value: */
void* create_new_integer(const int x){
  void* object=allocate_integer();

  assign_integer(object,x);
  return object;
}

/* Create an object, initialize it and push on the constants stack of an LVM: */
void push_new_constant_integer(const int x, struct LVM* lvm){
  void* object=allocate_integer();

  assign_integer(object,x);
  push_constant_object(object,lvm);
}

/* Create an object, initialize it and push on the stack of an LVM: */
void push_new_float(const double x, struct LVM* lvm){
  void* object=allocate_float();

  assign_float(object,x);
  push_object(object,lvm);
}

/* Create an object, initialize it and push on the constants stack of an LVM: */
void push_new_constant_float(const double x, struct LVM* lvm){
  void* object=allocate_float();
  
  assign_float(object,x);
  push_constant_object(object,lvm);
}

/* Create an object, initialize it and push on the stack of an LVM: */
void push_new_string(const char *x, struct LVM* lvm){
  void* object=allocate_string();
  
  assign_string(object,x);
  push_object(object,lvm);
}

/* Create an object, initialize it and push on the constants stack of an LVM: */
void push_new_constant_string(const char *x, struct LVM* lvm){
  void* object=allocate_string();

  assign_string(object,x);
  push_constant_object(object,lvm);
}

/* Create an object and push on the stack of an LVM: */
void push_new_array(const size_t size,struct LVM* lvm){
  push_object(allocate_array(size),lvm);
}

/* Create an object and push on the constants stack of an LVM: */
void push_new_constant_array(const size_t size,struct LVM* lvm){
  push_constant_object(allocate_array(size),lvm);
}

void output_object(void* object);

/* Create a clone of an object, re-building its whole structure: */
void* clone_object_with_map(void* object, struct map* m){
  void* r;
  void** already_cloned_r;

  /* Check for NULL objects: */
  if(object==NULL)
    return NULL;
  
  /* If we arrived here then object is not NULL. */
  
  /* If we already cloned this object which appears more than once in the
     structure then return the same pointer: */
  if((already_cloned_r=(void**)access_map(m,&object))!=NULL){
    printf("<Object ");
    output_object(object);
    printf(" HAS already been cloned.>\n");
    return *already_cloned_r;
  }
  else /* this is the first time this object appears in the structure; remember
          it: */
  insert_into_map(m,&object,&r);

  switch(((struct base_object*)object)->type){
    case integer_type:{
      r=allocate_integer();
      assign_integer(r,*integer2value(object));
      break;
    }
    case float_type:{
      r=allocate_float();
      assign_float(r,*float2value(object));
      break;
    }
    case string_type:{
      r=allocate_string();
      assign_string(r,string2value(object));
      break;
    }
    case array_type:{
      int i;

      r=allocate_array(((struct array_object*)object)->size);

      for(i=0;i<((struct array_object*)object)->size;i++)
        //if(((struct array_object*)object)->value[i] != NULL){
          ((struct array_object*)r)->value[i]=
             clone_object_with_map(((struct array_object*)object)->value[i],m);
          pin_object(((struct array_object*)object)->value[i]);
	//} /* if */
      break;
    }
  } /* switch */

  /*
  printf("<Now I have cloned ");
  output_object(object);
  printf(" into ");
  output_object(r);
  printf(">\n");
  */
  return r;
}

/* Create a clone of an object, re-building its whole structure: */
void* clone_object(void* object){
  void *clone;

  struct map* m=create_map(pointer_compare,
                           sizeof(void*),
                           sizeof(void*));
  clone=clone_object_with_map(object,m);
  destroy_map(m);

  return clone;
}

int output_object_fast(void* object){
  if(object==NULL){
    printf("nil");
    return 0;
  }

  switch(((struct base_object*)object)->type){
    case integer_type:{
      printf("%i",((struct integer_object*)object)->value);
      break;
    }
    case float_type:{
      printf("%lf",((struct float_object*)object)->value);
      break;
    }
    case string_type:{
      printf("%s",((struct string_object*)object)->value);
      break;
    }
    case array_type:{
      int i;

      printf("[");
      for(i=0;i<((struct array_object*)object)->size;i++){
        output_object_fast(((struct array_object*)object)->value[i]);
        if(i<((struct array_object*)object)->size - 1)
          printf(" ");
      }
      printf("]");
      break;
    }
    default:{
      fprintf(stderr,"output_object(): This cannot happen!\n");
      exit(EXIT_FAILURE);
    }
  } /* switch */
  return 0;
}

int output_object_with_map(void* object, struct map* m){
  if(object==NULL){
    printf("nil");
    return 0;
  }

  if(access_map(m,&object)!=NULL){
    printf("...");
    return -1;
  }
  else
    /* Remember we are writing this object: */
    insert_into_map(m,&object,&object);

  switch(((struct base_object*)object)->type){
    case integer_type:{
#ifdef OUTPUT_PINS_NUMBER
      printf("%i<%i>",((struct integer_object*)object)->value,((struct base_object*)object)->pins_number);
#else
      printf("%i",((struct integer_object*)object)->value);
#endif /* #ifdef OUTPUT_PINS_NUMBER */
      break;
    }
    case float_type:{
#ifdef OUTPUT_PINS_NUMBER
      printf("%lf<%i>",((struct float_object*)object)->value,((struct base_object*)object)->pins_number);
#else
      printf("%lf",((struct float_object*)object)->value);
#endif /* #ifdef OUTPUT_PINS_NUMBER */
      break;
    }
    case string_type:{
#ifdef OUTPUT_PINS_NUMBER
      printf("%s<%i>",((struct string_object*)object)->value,((struct base_object*)object)->pins_number);
#else
      printf("%s",((struct string_object*)object)->value);
#endif /* #ifdef OUTPUT_PINS_NUMBER */
      break;
    }
    case array_type:{
      int i;

      printf("[");
      for(i=0;i<((struct array_object*)object)->size;i++){
        output_object_with_map(((struct array_object*)object)->value[i],m);
        if(i<((struct array_object*)object)->size - 1)
          printf(" ");
      }
#ifdef OUTPUT_PINS_NUMBER
      printf("]<%i>",((struct base_object*)object)->pins_number);
#else
      printf("]");
#endif /* #ifdef OUTPUT_PINS_NUMBER */
      break;
    }
    default:{
      fprintf(stderr,"output_object(): This cannot happen!\n");
      exit(EXIT_FAILURE);
    }
  } /* switch */
  return 0;
}

void output_object(void* object){
  struct map* m=create_map(pointer_compare,
                           sizeof(void*),
                           sizeof(void*) /* unused */);
  output_object_with_map(object,m);
  destroy_map(m);
}

void output_list_fast(void *object){
  int i;
  void* x;

  if(object==NULL){
    printf("nil");
    return;
  }

  printf("(");
  x=object;
  while(x!=NULL){
    output_object_fast(((struct array_object*)x)->value[0]);

    x=((struct array_object*)x)->value[1];
    if(x!=NULL)
      printf(" ");
  }
  printf(")");
}

void output_list(void *object){
  int i;
  void* x;
  struct map* m;

  if(object==NULL){
    printf("nil");
    return;
  }
  m=create_map(pointer_compare,
	       sizeof(void*),
	       sizeof(void*) /* unused */);

  printf("(");
  x=object;
  while(x!=NULL){
    if(output_object_with_map(((struct array_object*)x)->value[0],m)==-1)
      break;

    x=((struct array_object*)x)->value[1];
    if(x!=NULL)
      printf(" ");
  }
  printf(")");
  destroy_map(m);
}

void read_constants(FILE *f,struct LVM *lvm){
  int x;
  
  do{
    fread(&x,sizeof(int),(size_t)1,f);

    if(feof(f))
      break;

    switch(x){
      case integer_type:{
        int parameter;

        fread(&parameter,sizeof(int),(size_t)1,f);
        push_new_constant_integer(parameter,lvm);
        break;
      }
      case float_type:{
        double parameter;
      
        fread(&parameter,sizeof(double),(size_t)1,f);
        push_new_constant_float(parameter,lvm);
        break;
      }
      case string_type:{
        int dimension;
        char *s;

        fread(&dimension,sizeof(int),(size_t)1,f);
        s=(char*)safe_malloc((dimension)*sizeof(char));
        fread(s,sizeof(char),(size_t)(dimension-1),f);
        s[dimension-1]='\0';
        push_new_constant_string(s,lvm);
        free(s);
        break;
      }
      case array_type:{
        int parameter;
        int i;

        fread(&parameter,sizeof(int),(size_t)1,f);
        push_new_constant_array((size_t)parameter,lvm);
        for(i=0;i<parameter;i++)
          ((struct array_object*)(lvm->stack[lvm->stack_pointer-1]))->value[i]=NULL;
        break;
      }
      default:{
        printf("read_constants(): This cannot happen!\n");
        exit(EXIT_FAILURE);
      }
    } /* switch */
  }while(!feof(f));
}

void read_instructions(FILE *f,struct LVM *lvm){
  int opcode; int argument;
  long current_position=ftell(f);
  long end_position;
  long i;

  /* Compute the number of instructions from the size of the file: */
  current_position=ftell(f);
  fseek(f,0,SEEK_END);
  end_position=ftell(f);
  fseek(f,current_position,SEEK_SET);
  lvm->instructions_number=
             (end_position-current_position)/sizeof(struct LVM_instruction);

  /* Allocate storage and read instructions: */
  lvm->instructions=(struct LVM_instruction*)
           safe_malloc(sizeof(struct LVM_instruction)*lvm->instructions_number);
  lvm->labels2instruction_number=(int*)
         safe_malloc(sizeof(int)*lvm->instructions_number);
  for(i=0;;i++){
    fread(&(lvm->instructions[i]),sizeof(struct LVM_instruction),(size_t)1,f);
    if(feof(f)||lvm->instructions[i].opcode==END_OF_INSTRUCTIONS)
      break;
    if(lvm->instructions[i].opcode==LABEL){
//      fprintf(stderr,"\t%i:\n",lvm->instructions[i].argument);

      /* Remember that this label will be bound to the istruction to be
         loaded at the i-th place: */
      lvm->labels2instruction_number[lvm->instructions[i].argument]=i;
//      fprintf(stderr,"Label %i: -> instruction %i\n",lvm->instructions[i].argument,i);

      i--; /* This is not a real instruction */
    }
/*
    else
      fprintf(stderr,"%i.\t%i %i\n",i,
                                    lvm->instructions[i].opcode,
                                    lvm->instructions[i].argument);
*/
  }
  /* Resize the instructions' array not to include 'LABEL'
     pseudo-instructions: */
  lvm->instructions_number=i;
  lvm->instructions=
         (struct LVM_instruction*)safe_realloc(lvm->instructions,
                                          i*sizeof(struct LVM_instruction));
//  fprintf(stderr,"[Numero di istruzioni: %i]\n",lvm->instructions_number);
}

void fix_labels(struct LVM *lvm){
  int i;
  
  /* Scan the instructions and replace label identifiers to instruction
     numbers: */
  for(i=0;i<lvm->instructions_number;i++)
    switch(lvm->instructions[i].opcode){
      case JUMP: case JUMP_IF_ZERO: case JUMP_IF_NONZERO:
      case LOAD_LABEL: case CLOSE: case CLOSE_RECURSIVELY: {
        /* This is a instruction with a label as parameter: store the 
           instruction number where to jump: */
        lvm->instructions[i].argument=
              lvm->labels2instruction_number[lvm->instructions[i].argument]-1;
        break;
      }
      default:{ /* This is not a jump instruction: do nothing. */ }
    } /* switch */

  /* Now free lvm->labels2instruction_number: it won't be used any more. */
  free(lvm->labels2instruction_number);
}

void* allocate_array(size_t size);
void push_object(void *object, struct LVM *lvm);
void push_new_integer(int value, struct LVM *lvm);

void create_first_frame(struct LVM *lvm){
  struct array_object *environment;
  int i;

  /* Just create a null pointer to a non-existing non-local environment: */
  environment=allocate_array(1);
  environment->value[0]=NULL;
  push_object(environment,lvm);

  /* Create a fake old frame pointer, with some special value: it would be 
     an error to use this just for anything since there is no frame under 
     this one. */
  push_new_integer(-1,lvm);
}

/* Returns NULL on error */
struct LVM* load_LVM(const char *file_name){
  FILE *f=fopen(file_name,"r");
  struct LVM *lvm=(struct LVM*)safe_malloc(sizeof(struct LVM));
  char heading[19];

#ifdef DEBUG
  fprintf(stderr,"Loading LVM module... ");
#endif
  if(f==0)
    return NULL;

  /* Read heading: */
  fread(heading,sizeof(char),(size_t)18,f);
  heading[18]='\0';
  if(strcmp(heading,"LVM bytecode v" BYTECODE_VERSION))
    return NULL; /* f is not a valid bytecode file */

  /* Set stack pointer: */
  lvm->stack_pointer=0;
  lvm->stack_size=BLOCK_SIZE;
  lvm->stack=(void**)safe_malloc(sizeof(void*)*lvm->stack_size);

  /* Set constants stack pointer: */
  lvm->constants_stack_pointer=0;
  lvm->constants_stack_size=BLOCK_SIZE;
  lvm->constants_stack=(void**)safe_malloc(sizeof(void*)*lvm->constants_stack_size);

  /* Read instructions: */
  read_instructions(f,lvm);

#ifdef DEBUG
  fprintf(stderr,"done; %i instructions.\n",lvm->instructions_number);
#endif

  /* Read constants: */
  read_constants(f,lvm);

  /* Close the file stream: */
  fclose(f);

  /* Set frame pointer: */
  lvm->frame_pointer=lvm->stack_pointer;

  /* Fix labels: */
  fix_labels(lvm);

  /* Create the first frame: */
  create_first_frame(lvm);

  /* Return the loaded and set-up machine: */
  return lvm;
}

void output_stack(struct LVM* lvm);

void destroy_LVM(struct LVM* lvm){
  /* Unpin everything is still pointed by the stack: */
#ifdef DEBUG
  fprintf(stderr,"Unpinning objects... ");
#endif
  for(lvm->stack_pointer--;lvm->stack_pointer>=0;lvm->stack_pointer--)
    unpin_object(lvm->stack[lvm->stack_pointer]);
#ifdef DEBUG
  fprintf(stderr,"done\n");
#endif
  /* Free the stack: */
  free(lvm->stack);

  /* Unpin everything is still pointed by the constants stack: */
#ifdef DEBUG
  fprintf(stderr,"Unpinning constant objects... ");
#endif
  for(lvm->constants_stack_pointer--;lvm->constants_stack_pointer>=0;
      lvm->constants_stack_pointer--)
    unpin_object(lvm->constants_stack[lvm->constants_stack_pointer]);
#ifdef DEBUG
  fprintf(stderr,"done\n");
#endif

  /* Free the constants stack: */
  free(lvm->constants_stack);

  /* Free the instructions: */
  free(lvm->instructions);

  /* Free the structure: */
  free(lvm);
}

/* Increment a stack pointer resizing the stack if needed: */
void increment_stack_pointer(void*** stack, int* stack_pointer, 
                             int* stack_size){
#ifdef DYNAMIC_STACK
  /* Resize the stack if needed: */
  if((*stack_pointer) == (*stack_size)){
#ifdef STACK_RESIZE_LOG
    fprintf(stdout,"[DEVO INGRANDIRE LO STACK: %i -> %i]\n",
            *stack_size,
            (*stack_size)+BLOCK_SIZE);
#endif
    *stack_size+=BLOCK_SIZE;
    *stack=(void**)safe_realloc(*stack,sizeof(void*)*(size_t)(*stack_size));
  }

#elif defined(STACK_OVERFLOW_CHECK)
  /* Test the overflow condition: */
  if((*stack_pointer) == (*stack_size)){
    fprintf(stderr,"Stack overflow\n");
    exit(EXIT_FAILURE);
  }
#endif

  /* Increment the stack pointer: */
  (*stack_pointer)++;
}

/* Push an existing object on the stack and add a pin to it: */
void push_object(void *object,struct LVM *lvm){
  increment_stack_pointer(&lvm->stack, &lvm->stack_pointer, &lvm->stack_size);
  pin_object(object);
  lvm->stack[lvm->stack_pointer-1]=object;
}

/* Push an existing constant object on the constant stack and add a pin to 
   it: */
void push_constant_object(void *object,struct LVM *lvm){
  increment_stack_pointer(&lvm->constants_stack, 
                          &lvm->constants_stack_pointer, 
                          &lvm->constants_stack_size);
  pin_object(object);
  lvm->constants_stack[lvm->constants_stack_pointer-1]=object;
}

void* top(struct LVM* lvm){
  return lvm->stack[lvm->stack_pointer-1];
}

void* under_top(struct LVM* lvm){
  return lvm->stack[lvm->stack_pointer-2];
}

void* under_under_top(struct LVM* lvm){
  return lvm->stack[lvm->stack_pointer-3];
}

int unary_minus(int x){
  return -x;
}

double unary_minus_float(double x){
  return -x;
}

void integer_to_integer_operator(int (*f)(int),struct LVM* lvm){
  int x=f(*integer2value(top(lvm)));
  pop(lvm);
  push_new_integer(x,lvm);
}

void float_to_float_operator(double (*f)(double),struct LVM* lvm){
  double x=f(*float2value(top(lvm)));
  pop(lvm);
  push_new_float(x,lvm);
}

void integer_to_float_operator(double (*f)(int),struct LVM* lvm){
  double x=f(*integer2value(top(lvm)));
  pop(lvm);
  push_new_float(x,lvm);
}

void float_to_integer_operator(int (*f)(double),struct LVM* lvm){
  int x=f(*float2value(top(lvm)));
  pop(lvm);
  push_new_integer(x,lvm);
}

int sum(int x,int y){
  return x+y;
}
int minus(int x,int y){
  return x-y;
}
int times(int x,int y){
  return x*y;
}
int divided(int x,int y){
#ifdef RUNTIME_SANITY_CHECKS
  if(y!=0)
#endif
    return x/y;
#ifdef RUNTIME_SANITY_CHECKS
  else
    fatal("integer division by zero");
#endif
}
int modulo(int x,int y){
#ifdef RUNTIME_SANITY_CHECKS
  if(y!=0)
#endif
    return x%y;
#ifdef RUNTIME_SANITY_CHECKS
  else
    fatal("modulo by zero");
#endif
}
int power(int x,int y){
  int i,r=1;
  for(i=0;i<y;i++)
    r*=x;
  return r;
}
double sum_float(double x,double y){
  return x+y;
}
double minus_float(double x,double y){
  return x-y;
}
double times_float(double x,double y){
  return x*y;
}
double divided_float(double x,double y){
#ifdef RUNTIME_SANITY_CHECKS
  if(y!=0.0)
#endif
    return x/y;
#ifdef RUNTIME_SANITY_CHECKS
  else
    fatal("float division by zero");
#endif
}
double power_float(double x,double y){
  return pow(x,y);
}

int equal(int x,int y){
  return x==y;
}
int not_equal(int x,int y){
  return x!=y;
}
int less_than(int x,int y){
  return x<y;
}
int greater_than(int x,int y){
  return x>y;
}
int less_than_or_equal(int x,int y){
  return x<=y;
}
int greater_than_or_equal(int x,int y){
  return x>=y;
}

int equal_float(double x,double y){
  return x==y;
}
int not_equal_float(double x,double y){
  return x!=y;
}
int less_than_float(double x,double y){
  return x<y;
}
int greater_than_float(double x,double y){
  return x>y;
}
int less_than_or_equal_float(double x,double y){
  return x<=y;
}
int greater_than_or_equal_float(double x,double y){
  return x>=y;
}

int equal_string(char* x, char* y){
  return !strcmp(x,y);
}

int not_equal_string(char* x, char* y){
  return strcmp(x,y);
}

int less_than_string(char* x, char* y){
  return strcmp(x,y) < 0;
}

int greater_than_string(char* x, char* y){
  return strcmp(x,y) > 0;
}

int less_than_or_equal_string(char* x, char* y){
  return strcmp(x,y) <= 0;
}

int greater_than_or_equal_string(char* x, char* y){
  return strcmp(x,y) >= 0;
}

int zero(int x){
  return !x;
}
int nonzero(int x){
  return x;
}
int not_function(int x){
  return !x;
}
int and_function(int x,int y){
  return x && y;
}
int or_function(int x,int y){
  return x || y;
}

int cast_float_to_integer(double x){
  return (int)x;
}

double cast_integer_to_float(int x){
  return (double)x;
}

void integer2_to_integer_operator(int (*f)(int,int),struct LVM* lvm){
  int x=f(*integer2value(under_top(lvm)),
          *integer2value(top(lvm)));
  pop(lvm);
  pop(lvm);
  push_new_integer(x,lvm);
}

void integer_immediate_to_integer_operator(int (*f)(int,int),struct LVM* lvm){
  int x=f(*integer2value(top(lvm)),
          lvm->instructions[lvm->instruction_pointer].argument);
  unpin_object(top(lvm));
  lvm->stack[lvm->stack_pointer-1]=create_new_integer(x);
  pin_object(lvm->stack[lvm->stack_pointer-1]);
}

void float2_to_float_operator(double (*f)(double,double),struct LVM* lvm){
  double x=f(*float2value(under_top(lvm)),
             *float2value(top(lvm)));
  pop(lvm);
  pop(lvm);
  push_new_float(x,lvm);
}

void float2_to_integer_operator(int (*f)(double,double),struct LVM* lvm){
  int x=f(*float2value(under_top(lvm)),
          *float2value(top(lvm)));
  pop(lvm);
  pop(lvm);
  push_new_integer(x,lvm);
}

void string2_to_integer_operator(int (*f)(char*,char*),struct LVM* lvm){
  int x=f(string2value(under_top(lvm)),
          string2value(top(lvm)));
  pop(lvm);
  pop(lvm);
  push_new_integer(x,lvm);
}

void conditional(int (*f)(int),struct LVM* lvm,int instruction){
  int x=f(*integer2value(top(lvm)));
  pop(lvm);
  if(x)
    lvm->instruction_pointer=instruction;
}

void output_stack(struct LVM* lvm){
  int i;
  for(i=lvm->stack_pointer-1;i>=0;i--){
    printf("S%2i || F%2i -> %9p = ", i,i-lvm->frame_pointer, lvm->stack[i]);
    output_object(lvm->stack[i]);
/*/
    if(lvm->stack[i] != NULL)
      if(is_a_circular_closure(lvm->stack[i]))
        printf("CC");
/**/
    printf("\n");
  }
}

/* This only assumes x is a non-null LVM object. */
int is_part_of_a_circular_closure(struct array_object *x){
  if(((struct base_object*)x)->type != array_type)
    return 0;
  else if(x->size != 2)
    return 0;
  else if(x->value[1]==NULL)
    return 0;
  else if(((struct base_object*)(x->value[1]))->type != array_type)
    return 0;
  else if(((struct array_object*)(x->value[1]))->size != 2)
    return 0;
  else 
    /* if a cycle exists then x is a circular closure; the LVM only 
       admits this kind of circularity. If the cycle does not exist
       then x is not part of a closure or the closure is not
       circular. */
    return ((struct array_object*)(x->value[1]))->value[1] == x;
}

/* This assumes x is one or the other part of a circular closure. */
int is_the_environment_part_of_a_circular_closure(struct array_object *x){
    return ((struct array_object*)(x->value[0]))->type == array_type;
}

/* This assumes x is one or the other part of a circular closure. */
int is_the_closure_part_of_a_circular_closure(struct array_object *x){
  return ((struct array_object*)(x->value[0]))->type == integer_type;
}

/* This assumes closure is an actual non-null closure. */
int is_a_circular_closure(struct array_object *closure){
  if(((struct array_object*)(closure->value[1]))->size != 2)
    return 0;
  else
    return closure == ((struct array_object*)(closure->value[1]))->value[1];
}

  #define IP            (lvm->instruction_pointer)
  #define OPCODE        (lvm->instructions[IP].opcode)
  #define ARGUMENT      (lvm->instructions[IP].argument)
  #define SP            (lvm->stack_pointer)
  #define FP            (lvm->frame_pointer)
  #define TOP           (lvm->stack[SP-1])
  #define UNDERTOP      (lvm->stack[SP-2])
  #define UNDERUNDERTOP (lvm->stack[SP-3])

void push_frame(int arguments_no,struct LVM *lvm){
  int i;
  struct array_object* array=allocate_array(arguments_no+1);

  array->value[0]=lvm->stack[FP];
  pin_object(array->value[0]);
  for(i=0;i<arguments_no;i++)
    array->value[i+1]=lvm->stack[SP-arguments_no+i];
  SP-=(arguments_no-1);
  pin_object(array);
  TOP=array;
  push_new_integer(FP,lvm);
  FP=SP-2;
}

void pop_frame(int is_there_a_return_value, struct LVM* lvm){
  /* Remember the old frame pointer before unpinning it: */
  int old_frame_pointer=((struct integer_object*)(lvm->stack[FP+1]))->value;

  /* Remember and pin the object to return, if needed: */
  void* returned_value;
  if(is_there_a_return_value){
    returned_value=TOP;
    pin_object(returned_value);
  }

  /* Destroy everything we don't need on the stack: */
  while(SP > FP)
    pop(lvm);

  /* Push and unpin the object to return, if needed: */
  if(is_there_a_return_value){
    push_object(returned_value,lvm);
    unpin_object(returned_value);
  }

  /* Restore frame_pointer: */
  lvm->frame_pointer=old_frame_pointer;
}

void pop_call_frame(int is_there_a_return_value,struct LVM* lvm){
  /* Remember the old instruction pointer before unpinning it: */ 
  int old_instruction_pointer=((struct integer_object*)(lvm->stack[FP+2]))->value;

  /* Pop this frame: */
  pop_frame(is_there_a_return_value,lvm);

  /* Restore instruction pointer: */
  lvm->instruction_pointer=old_instruction_pointer;
}

int run_LVM(struct LVM *lvm){
  for(IP=0; IP < lvm->instructions_number; IP++){
#ifdef TRACE
    output_stack(lvm);
    printf("\n<Showing the stack updated by instruction %i:>\n",IP);
#endif
    switch(OPCODE){
      case INPUT_STRING:{
        char *s=input_a_line();
        push_new_string(s,lvm);
        free(s);
        
        break;
      }

      case INPUT_INTEGER:{
        char *s;
        int i;
        int success;

        do{
          s=input_a_line();
          success=(sscanf(s,"%i",&i)==1);
          free(s);
          if(!success)
            printf("\aYou should enter an integer number. Please try again:\n>");
        }while(!success);
        push_new_integer(i,lvm);

        break;
      }

      case INPUT_FLOAT:{
        char *s;
        double f;
        int success;

        do{
          s=input_a_line();
          success=(sscanf(s,"%lf",&f)==1);
          free(s);
          if(!success)
            printf("\aYou should enter a floating-point number. Please try again:\n>");
        }while(!success);
        push_new_float(f,lvm);

        break;
      }

      case ARE_THE_SAME_OBJECT:{
        int are_the_same_object=(TOP==UNDERTOP);

        pop(lvm);
        unpin_object(TOP);
        TOP=create_new_integer(are_the_same_object);
        pin_object(TOP);
        break;
      }

      case ARE_NOT_THE_SAME_OBJECT:{
        int are_not_the_same_object=(TOP!=UNDERTOP);

        pop(lvm);
        unpin_object(TOP);
        TOP=create_new_integer(are_not_the_same_object);
        pin_object(TOP);
        break;
      }

      case CLOSE:{
	/*
        This version seems buggy; it has problems with the pins count.

        push_new_array(2,lvm);
        ((struct array_object*)TOP)->value[0]=create_new_integer(ARGUMENT);
        pin_object(((struct array_object*)TOP)->value[0]);
        ((struct array_object*)TOP)->value[1]=lvm->stack[FP];
        pin_object(((struct array_object*)TOP)->value[1]);
        */
	struct array_object *a=allocate_array(2);
        a->value[0]=create_new_integer(ARGUMENT);
        a->value[1]=lvm->stack[FP];
        pin_object(a->value[0]);
        pin_object(a->value[1]);
        push_object(a,lvm);
        break;
      }

      case CLOSE_RECURSIVELY:{
        struct array_object *closure=allocate_array(2);
        struct array_object *environment=allocate_array(2);

        closure->value[0]=create_new_integer(ARGUMENT);
        pin_object(closure->value[0]);
	closure->value[1]=environment;
        /* No need to pin pin_object(closure->value[1]): the pins' number on the environment part
           is not used. */
#ifdef DEBUG
        ((struct base_object*)(closure->value[1]))->pins_number = -10000; // an 'unusual' value, for tracing
#endif

        environment->value[0]=lvm->stack[FP];
        pin_object(environment->value[0]); // !!!!!!!!

        environment->value[1]=closure;
        /* I intentionally don't pin_object(environment->value[1]); this is the circular reference. */

        push_object(closure,lvm);

        //printf("\nCLOSE_RECURSIVELY:\t");output_object(closure);printf("\n");

        break;
      }

      case LOAD_NON_LOCAL_ENVIRONMENT:{
        int i;
        struct array_object* environment=lvm->stack[FP];

        for(i=0;i<ARGUMENT;i++)
          environment=(struct array_object*)(environment->value[0]);

        push_object(environment,lvm);

        break;
      }

      case LOAD_ARRAY_ELEMENT_IMMEDIATE:{
        void *element;

#ifdef DEBUG /* there's no point in doing this test on correct compiled code */
        if((ARGUMENT<0)||(ARGUMENT>=((struct array_object*)TOP)->size))
          fatal("out of bounds in LOAD_ARRAY_ELEMENT_IMMEDIATE");
#endif

        element=((struct array_object*)TOP)->value[ARGUMENT];
        
        /* replace array with element: */
        pin_object(element);
        unpin_object(TOP);
        TOP=element;
        
        break;
      }

      case LOAD_ARRAY_ELEMENT:{ // To do: ottimizzare...
        int index=((struct integer_object*)TOP)->value;
        void *element;

#ifdef RUNTIME_SANITY_CHECKS
        if((index<0)||(index>=((struct array_object*)UNDERTOP)->size))
	  fatal("out of bounds in LOAD_ARRAY_ELEMENT");
#endif

        element=((struct array_object*)UNDERTOP)->value[index];

        /* Keep the element we are going to push! */
        pin_object(element);

        /* Pop arguments: */
        pop(lvm);pop(lvm);

        /* Push result: */
        push_object(element,lvm);

        /* Remove the pin we added before; push_object() added anohter one. */
        unpin_object(element);

        break;
      }

      case INTEGER_TO_FLOAT:{
        integer_to_float_operator(cast_integer_to_float,lvm);
        break;
      }

      case FLOAT_TO_INTEGER:{
        float_to_integer_operator(cast_float_to_integer,lvm);
        break;
      }

      case PUSH_FRAME:{
        push_frame(ARGUMENT /* number of paramters/locals */,lvm);
        break;
      }

      case POP_FRAME:{
        pop_frame(0/* nothing to return */,lvm);
        break;
      }

      case POP_FRAME_AND_RETURN:{
        pop_frame(1/* something to return */,lvm);
        break;
      }

      case CALL_STACK:{
        /* The closure is at SP - ARGUMENT - 1; parameters are above */
        struct array_object* closure=lvm->stack[SP-ARGUMENT-1];
        struct array_object* new_local_environment=allocate_array(ARGUMENT+1);
        int where_to_jump=((struct integer_object*)(closure->value[0]))->value;
        int i;
        
        new_local_environment->value[0]=closure->value[1];
        for(i=0;i<ARGUMENT;i++)
          new_local_environment->value[i+1]=lvm->stack[SP-ARGUMENT+i];
        
        lvm->stack[SP-ARGUMENT-1]=new_local_environment;
        pin_object(new_local_environment);
        SP-=ARGUMENT;

        if(!is_a_circular_closure(closure)){
          pin_object(closure->value[1]);
          unpin_object(closure);
        } /* else don't touch pins: they are not meaningful in the environment part.*/

        /* Push the old frame pointer and the old instruction pointer: */
        push_new_integer(FP,lvm);
        push_new_integer(IP,lvm);
        
        /* Set frame pointer and instruction pointer: */
        FP=SP-3;
        IP=where_to_jump;

        break;
      }

      case RETURN:{
        pop_call_frame(0/* nothing to return */,lvm);

        break;
      }

      case RETURN_VALUE:{
        pop_call_frame(1/* something to return */,lvm);

        break;
      }

      case ACCESS_LOCAL:{
#ifdef DEBUG /* there's no point in doing this test on correct compiled code */
        if(ARGUMENT==0)
          fatal("ACCESS_LOCAL should have a non-zero parameter");
        if((ARGUMENT<0)||(ARGUMENT>=((struct array_object*)(lvm->stack[FP]))->size))
          fatal("out of bounds in ACCESS_LOCAL");
#endif
        push_object(((struct array_object*)(lvm->stack[FP]))->value[ARGUMENT],lvm);
        break;
      }

      case MAKE_ARRAY:{
        int i;
        struct array_object* array=allocate_array(ARGUMENT);

        for(i=0;i<ARGUMENT;i++)
          array->value[i]=lvm->stack[SP-ARGUMENT+i];
        SP-=ARGUMENT;
	push_object(array,lvm);
        break;
      }

      case CONCATENATE_STRINGS:{
        char* temp;
        struct string_object* s1=UNDERTOP;
        struct string_object* s2=TOP;

        temp=(char*)malloc(sizeof(char)*(s1->size+s2->size+1));
        strcpy(temp,s1->value);
        strcpy(temp+s1->size,s2->value);
        pop(lvm);pop(lvm);
        push_new_string(temp,lvm);
        free(temp);
        break;
      }
      case FROM_TO_STRING:{
	char* s=((struct string_object*)UNDERUNDERTOP)->value;
        int from=((struct integer_object*)UNDERTOP)->value;
        int to=((struct integer_object*)TOP)->value;
        char* temp;

#ifdef RUNTIME_SANITY_CHECKS
        if(from>to)
          fatal("lower bound greater than upper bound in FROM_TO_STRING");
	else if(from<0)
	  fatal("lower bound less than 0 in FROM_TO_STRING");
        else if(to>=((struct string_object*)UNDERUNDERTOP)->size)
          fatal("upper bound greater than or equal to size in FROM_TO_STRING");
#endif

        temp=(char*)malloc(sizeof(char)*(to-from+1));

        strncpy(temp,s+from,to-from+1);
        temp[to-from+1]='\0';
	/*
        strncpy(temp,s+from-1,to-from+1);
        temp[to-from+1]='\0';
	*/
        pop(lvm);pop(lvm);pop(lvm);
        push_new_string(temp,lvm);
        free(temp);
        break;
      }
      case SWAP:
      {
        void* temp;

        temp=TOP;
        TOP=UNDERTOP;
        UNDERTOP=temp;
        break;
      }
      case LENGTH_OF_STRING:{
	int size=((struct string_object*)TOP)->size;

        pop(lvm);
        push_new_integer(size,lvm);
        break;
      }
      case SIZE_OF_ARRAY:{
        int size=((struct array_object*)TOP)->size;
        
        pop(lvm);
        push_new_integer(size,lvm);
        break;
      }
      case RESIZE_ARRAY:{
        int old_size=((struct array_object*)UNDERTOP)->size;
        int new_size=((struct integer_object*)TOP)->value;
        void* array=UNDERTOP;
        int i;

#ifdef RUNTIME_SANITY_CHECKS
        if(new_size<0)
          fatal("size less than zero in RESIZE_ARRAY");
#endif
        /* Pop the integer parameter, leave alone the array: */
        pop(lvm);
        
        /* Resize the array: */
        if(new_size<old_size){
          for(i=new_size;i<old_size;i++)
            unpin_object(((struct array_object*)TOP)->value[i]);

          ((struct array_object*)TOP)->value=
                 	 (void**)safe_realloc(((struct array_object*)TOP)->value,
                                         sizeof(void*)*new_size);
	}
        else{
          ((struct array_object*)TOP)->value=
	           (void**)safe_realloc(((struct array_object*)TOP)->value,
           		           sizeof(void*)*new_size);
          for(i=old_size;i<new_size;i++)
            ((struct array_object*)TOP)->value[i]=NULL;
        }
        
        /* Update the 'size' field: */
        ((struct array_object*)TOP)->size=new_size;

        break;
      }

      case STORE:{
        /* Unpin the reference we're going to overwrite then overwrite it: */
        unpin_object(lvm->stack[FP+ARGUMENT]);
        lvm->stack[FP+ARGUMENT]=TOP;

        /* Add a pin to the object we are adding a pointer to: */
        pin_object(TOP);

        /* Pop the argument: */
        pop(lvm);
        break;
      }

      case LOAD:{
        push_object(lvm->stack[FP+ARGUMENT],
                    lvm);
        break;
      }

      case LOAD_LABEL:{
        push_new_integer(ARGUMENT,lvm);
        break;
      }

      case NIL:{
        push_object(NULL,lvm);
        break;
      }

      case IS_NULL:{
        void *top=TOP;

        pop(lvm); /* I don't mind if TOP is destroyed... */
        if(top==NULL)
          push_new_integer(1,lvm); /* push 'true' */
        else
          push_new_integer(0,lvm); /* push 'false' */
        break;
      }

      case IS_NOT_NULL:{
        void *top=TOP;

        pop(lvm); /* I don't mind if TOP is destroyed... */
        if(top!=NULL)
          push_new_integer(1,lvm); /* push 'true' */
        else
          push_new_integer(0,lvm); /* push 'false' */
        break;
      }

      case CONS:{
        void* cons=allocate_array(2);

        ((struct array_object*)cons)->value[0]=UNDERTOP;
        ((struct array_object*)cons)->value[1]=TOP;
        pin_object(TOP);
        pin_object(UNDERTOP);

        pop(lvm);
        pop(lvm);
        push_object(cons,lvm);

        break;
      }
       
      case CAR:{
        void* car;

#ifdef RUNTIME_SANITY_CHECKS
        if(TOP==NULL)
          fatal("CAR is applied to a NULL object");
#endif
        car=((struct array_object*)TOP)->value[0];
        pin_object(car);
        pop(lvm);
        push_object(car,lvm);
        unpin_object(car);
        break;
      }
      
      case CDR:{
        void* cdr;

#ifdef RUNTIME_SANITY_CHECKS
        if(TOP==NULL)
          fatal("CDR is applied to a NULL object");
#endif
        cdr=((struct array_object*)TOP)->value[1];
        pin_object(cdr);
        pop(lvm);
        push_object(cdr,lvm);
        unpin_object(cdr);
        break;
      }
      
      case RPLACA:{
#ifdef RUNTIME_SANITY_CHECKS
        if(UNDERTOP==NULL)
          fatal("RPLACA is applied to a NULL object");
#endif

        unpin_object(((struct array_object*)UNDERTOP)->value[0]);
        ((struct array_object*)UNDERTOP)->value[0]=TOP;
        pin_object(TOP);
        pop(lvm);
        pop(lvm);
        break;
      }
      
      case RPLACD:{
#ifdef RUNTIME_SANITY_CHECKS
        if(UNDERTOP==NULL)
          fatal("RPLACD is applied to a NULL object");
#endif

        unpin_object(((struct array_object*)UNDERTOP)->value[1]);
        ((struct array_object*)UNDERTOP)->value[1]=TOP;
        pin_object(TOP);
        pop(lvm);
        pop(lvm);
        break;
      }

      case STORE_ARRAY_ELEMENT:{
        int index=((struct integer_object*)UNDERTOP)->value;

#ifdef RUNTIME_SANITY_CHECKS
        if((index<0)||(index>=((struct array_object*)UNDERUNDERTOP)->size))
          fatal("out of bounds in STORE_ARRAY_ELEMENT");
#endif

        /* Unpin the current n-th element, insert the new one at the
           nth-place then pin it: */
        unpin_object(((struct array_object*)UNDERUNDERTOP)->value[index]);
        ((struct array_object*)UNDERUNDERTOP)->value[index]=TOP;
        pin_object(TOP);

        /* Pop arguments: */
        pop(lvm);
        pop(lvm);
        pop(lvm);

        break;
      }

      case CREATE_INTEGER:{
        push_new_integer(0,lvm);
        break;
      }
      case CREATE_FLOAT:{
        push_new_float(0.0,lvm);
        break;
      }
      case CREATE_STRING:{
        push_new_string("",lvm);
        break;
      }
      case CREATE_ARRAY:{
        int dimension=((struct integer_object*)TOP)->value;
        int i;
        struct array_object *a=allocate_array(dimension);

        for(i=0;i<dimension;i++)
          a->value[i]=NULL;
        pop(lvm);
        push_object(a,lvm);

        break;
      }

      case OUTPUT_FAST:{
        output_object_fast(TOP);
        pop(lvm);
        break;
      }
 
      case OUTPUT_LIST_FAST:{
        output_list_fast(TOP);
        pop(lvm);
        break;
      }
 
      case OUTPUT:{
#ifdef TRACE
        printf("OUTPUT: ");
#endif
        output_object(TOP);
	pop(lvm);
#ifdef TRACE
        printf("\n");
#endif
        break;
      }

      case OUTPUT_LIST:{
        output_list(TOP);
        pop(lvm);
        break;
      }

      case JUMP_IF_ZERO:{
        conditional(zero,lvm,ARGUMENT);
        break;
      }
      case JUMP_IF_NONZERO:{
        conditional(nonzero,lvm,ARGUMENT);
        break;
      }

      case AND:{
        integer2_to_integer_operator(and_function,lvm);
        break;
      }
      case OR:{
        integer2_to_integer_operator(or_function,lvm);
        break;
      }
      case NOT:{
        integer_to_integer_operator(not_function,lvm);
        break;
      }

      case PLUS:{
        integer2_to_integer_operator(sum,lvm);
        break;
      }
      case MINUS:{
        integer2_to_integer_operator(minus,lvm);
        break;
      }
      case TIMES:{
        integer2_to_integer_operator(times,lvm);
        break;
      }
      case DIVIDED:{
        integer2_to_integer_operator(divided,lvm);
        break;
      }
      case MODULO:{
        integer2_to_integer_operator(modulo,lvm);
        break;
      }
      case POWER:{
        integer2_to_integer_operator(power,lvm);
        break;
      }

      case PLUS_IMMEDIATE:{
        integer_immediate_to_integer_operator(sum,lvm);
        break;
      }
      case MINUS_IMMEDIATE:{
        integer_immediate_to_integer_operator(minus,lvm);
        break;
      }
      case TIMES_IMMEDIATE:{
        integer_immediate_to_integer_operator(times,lvm);
        break;
      }
      case DIVIDED_IMMEDIATE:{
        integer_immediate_to_integer_operator(divided,lvm);
        break;
      }
      case MODULO_IMMEDIATE:{
        integer_immediate_to_integer_operator(modulo,lvm);
        break;
      }
      case POWER_IMMEDIATE:{
        integer_immediate_to_integer_operator(power,lvm);
        break;
      }

      case EQUAL:{
        integer2_to_integer_operator(equal,lvm);
        break;
      }
      case NOT_EQUAL:{
        integer2_to_integer_operator(not_equal,lvm);
        break;
      }
      case LESS_THAN:{
        integer2_to_integer_operator(less_than,lvm);
        break;
      }
      case GREATER_THAN:{
        integer2_to_integer_operator(greater_than,lvm);
        break;
      }
      case LESS_THAN_OR_EQUAL:{
        integer2_to_integer_operator(less_than_or_equal,lvm);
        break;
      }
      case GREATER_THAN_OR_EQUAL:{
        integer2_to_integer_operator(greater_than_or_equal,lvm);
        break;
      }

      case EQUAL_FLOAT:{
        float2_to_integer_operator(equal_float,lvm);
        break;
      }
      case NOT_EQUAL_FLOAT:{
        float2_to_integer_operator(not_equal_float,lvm);
        break;
      }
      case LESS_THAN_FLOAT:{
        float2_to_integer_operator(less_than_float,lvm);
        break;
      }
      case GREATER_THAN_FLOAT:{
        float2_to_integer_operator(greater_than_float,lvm);
        break;
      }
      case LESS_THAN_OR_EQUAL_FLOAT:{
        float2_to_integer_operator(less_than_or_equal_float,lvm);
        break;
      }
      case GREATER_THAN_OR_EQUAL_FLOAT:{
        float2_to_integer_operator(greater_than_or_equal_float,lvm);
        break;
      }
      case EQUAL_STRING:{
        string2_to_integer_operator(equal_string,lvm);
        break;
      }
      case NOT_EQUAL_STRING:{
        string2_to_integer_operator(not_equal_string,lvm);
        break;
      }
      case LESS_THAN_STRING:{
        string2_to_integer_operator(less_than_string,lvm);
        break;
      }
      case GREATER_THAN_STRING:{
        string2_to_integer_operator(greater_than_string,lvm);
        break;
      }
      case LESS_THAN_OR_EQUAL_STRING:{
        string2_to_integer_operator(less_than_or_equal_string,lvm);
        break;
      }
      case GREATER_THAN_OR_EQUAL_STRING:{
        string2_to_integer_operator(greater_than_or_equal_string,lvm);
        break;
      }
      //////////////////////////////////////////////
      case EQUAL_GENERIC:{
        switch(((struct base_object*)TOP)->type){
   	  case float_type:{
            float2_to_integer_operator(equal_float,lvm);
	    break;
	  }
  	  case string_type:{
            string2_to_integer_operator(equal_string,lvm);
            break;
          }
	  default:{ 
            /* I compare any two objects else as integers; */
            integer2_to_integer_operator(equal,lvm);
            break;
          }
	} /* switch */
        break;
      }
      case NOT_EQUAL_GENERIC:{
        switch(((struct base_object*)TOP)->type){
  	  case float_type:{
	    float2_to_integer_operator(not_equal_float,lvm);
	    break;
	  }
	  case string_type:{
	    string2_to_integer_operator(not_equal_string,lvm);
	    break;
	  }
	  default:{
	    /* I compare any two objects else as integers; */
	    integer2_to_integer_operator(not_equal,lvm);
            break;
	  }
        } /* switch */
        break;
      }
      case LESS_THAN_GENERIC:{
        switch(((struct base_object*)TOP)->type){
	  case float_type:{
	    float2_to_integer_operator(less_than_float,lvm);
	    break;
	  }
	  case string_type:{
	    string2_to_integer_operator(less_than_string,lvm);
	    break;
	  }
	  default:{
	    /* I compare any two objects else as integers; */
	    integer2_to_integer_operator(less_than,lvm);
            break;
	  }
        } /* switch */
        break;
      }
      case GREATER_THAN_GENERIC:{
        switch(((struct base_object*)TOP)->type){
          case float_type:{
            float2_to_integer_operator(greater_than_float,lvm);
            break;
          }
          case string_type:{
            string2_to_integer_operator(greater_than_string,lvm);
            break;
          }
          default:{
            /* I compare any two objects else as integers; */
            integer2_to_integer_operator(greater_than,lvm);
            break;
          }
        } /* switch */
        break;
      }
      case LESS_THAN_OR_EQUAL_GENERIC:{
        switch(((struct base_object*)TOP)->type){
          case float_type:{
            float2_to_integer_operator(less_than_or_equal_float,lvm);
            break;
          }
          case string_type:{
            string2_to_integer_operator(less_than_or_equal_string,lvm);
            break;
          }
          default:{
            /* I compare any two objects else as integers; */
            integer2_to_integer_operator(less_than_or_equal,lvm);
            break;
          }
        } /* switch */
        break;
      }
      case GREATER_THAN_OR_EQUAL_GENERIC:{
        switch(((struct base_object*)TOP)->type){
          case float_type:{
            float2_to_integer_operator(greater_than_or_equal_float,lvm);
            break;
          }
          case string_type:{
            string2_to_integer_operator(greater_than_or_equal_string,lvm);
            break;
          }
          default:{
            /* I compare any two objects else as integers; */
            integer2_to_integer_operator(greater_than_or_equal,lvm);
            break;
          }
        } /* switch */
        break;
      }
      //////////////////////////////////////////////
      case UNARY_MINUS:{
        integer_to_integer_operator(unary_minus,lvm);
        break;
      }

      case PLUS_FLOAT:{
        float2_to_float_operator(sum_float,lvm);
        break;
      }
      case MINUS_FLOAT:{
        float2_to_float_operator(minus_float,lvm);
        break;
      }
      case TIMES_FLOAT:{
        float2_to_float_operator(times_float,lvm);
        break;
      }
      case DIVIDED_FLOAT:{
        float2_to_float_operator(divided_float,lvm);
        break;
      }
      case POWER_FLOAT:{
        float2_to_float_operator(power_float,lvm);
        break;
      }
      case UNARY_MINUS_FLOAT:{
        float_to_float_operator(unary_minus_float,lvm);
        break;
      }

      case ASSIGN:{
        assign_object(UNDERTOP,TOP);
        pop(lvm);
        pop(lvm);
        break;
      }
      
      case LOAD_ABSOLUTE:{
        push_object(lvm->stack[ARGUMENT],lvm);
        break;
      }
      
      case STORE_ABSOLUTE:{
        void* object=TOP;
        pin_object(object);
        pop(lvm);
        unpin_object(lvm->stack[ARGUMENT]);
        lvm->stack[ARGUMENT]=object;
        break;
      }
      
      case LOAD_CONSTANT:{
        //push_object(clone_object(lvm->constants_stack[ARGUMENT]), lvm);
        //unpin_object(TOP);
        push_object(lvm->constants_stack[ARGUMENT], lvm);
        break;
      }

      case COPY:{
        push_object(TOP,lvm);
        break;
      }

      case CLONE:{
        void* cloned_object=clone_object(TOP);

        pop(lvm);
        push_object(cloned_object,lvm);
        break;
      }

      case POP:{
        pop(lvm);
        break;
      }

      case JUMP:{
        IP=ARGUMENT;
        break;
      }

      case NOP:{
        /* Do nothing*/
        break;
      }

      case EXIT:{
        int return_code=((struct integer_object*)TOP)->value;

        pop(lvm);
        return return_code;
      }

      default:{
        fprintf(stderr,"Instruction %i: operation not implemented\n",
                lvm->instruction_pointer);
        exit(EXIT_FAILURE);
      }
    } /* switch */
//    fprintf(stderr,"<Done with instrucion %i>\n",IP);
  } /* for */

  /* We have not found an EXIT instruction yet. */
  
  fprintf(stderr,"End of instructions: no EXIT opcode found.\n");
  return -1;
}

int main(){
  struct LVM* lvm=load_LVM("x");
  int return_code=run_LVM(lvm);

#ifndef DEBUG
  if(return_code!=0)
#endif
    fprintf(stderr,"\n[Return code: %i]\n",return_code);

  destroy_LVM(lvm);
}






